from typing import Dict, Optional
from datetime import datetime

from xtb_broker.component.frozen import Frozen
from xtb_broker.component.constants import WEEKDAY
from xtb_broker.models.timetable import Timetable
from xtb_broker.models.shift import Shift

from xtb_broker.xtb import Xtb


class Symbol(Frozen):
    def __init__(self, xtb: Xtb, symbol: str):
        symbol_dict: Dict = xtb.get_client().get_symbol(symbol=symbol)
        self.__symbol: str = symbol_dict['symbol']
        self.__currency: str = symbol_dict['currency']
        self.__currency_profit: str = symbol_dict['currencyProfit']
        self.__contract_size: str = symbol_dict['contractSize']
        self.__precision: str = symbol_dict['precision']
        self.__bid: str = symbol_dict['bid']
        self.__ask: str = symbol_dict['ask']
        self.__swap_long: str = symbol_dict['swapLong']
        self.__swap_short: str = symbol_dict['swapShort']
        self.__spread: float = symbol_dict['spreadRaw']

        self.__time: int = symbol_dict['time']
        self.__pips_precision: int = symbol_dict['pipsPrecision']
        self.__tick_size: float = symbol_dict['tickSize']
        self.__leverage: float = symbol_dict['leverage']

        self.__timetable: Timetable = xtb.get_symbol_timetable(symbol=symbol)

    @property
    def symbol(self) -> str:
        return self.__symbol

    @property
    def currency(self) -> str:
        return self.__currency

    @property
    def currency_profit(self) -> str:
        return self.__currency_profit

    @property
    def contract_size(self) -> str:
        return self.__contract_size

    @property
    def precision(self) -> str:
        return self.__precision

    @property
    def bid(self) -> str:
        return self.__bid

    @property
    def ask(self) -> str:
        return self.__ask

    @property
    def swap_long(self) -> str:
        return self.__swap_long

    @property
    def swap_short(self) -> str:
        return self.__swap_short

    @property
    def timetable(self) -> Timetable:
        return self.__timetable

    @property
    def spread(self) -> float:
        return self.__spread

    @property
    def time(self) -> int:
        return self.__time

    @property
    def pips_precision(self) -> int:
        return self.__pips_precision

    @property
    def tick_size(self) -> float:
        return self.__tick_size

    @property
    def leverage(self) -> float:
        return self.__leverage

    def get_active_shift(self) -> Optional[Shift]:
        for shift in getattr(self.timetable, WEEKDAY[datetime.today().isoweekday()]):
            if shift.from_ts < datetime.now() < shift.to_ts:
                return shift



from typing import List, Dict, Optional
from datetime import datetime
from retrying import retry

from xtb_broker.component.exception import ArbitragePartiallyClosedException, TradeNotFoundException, \
    retry_if_arbitrage_partially_closed_exception
from xtb_broker.component.frozen import Frozen
from xtb_broker.component.constants import Cmd
from xtb_broker.component.logger import logger
from xtb_broker.models.position import Position
from xtb_broker.models.symbol import Symbol
from xtb_broker.models.shift import Shift
from xtb_broker.models.trade import Trade
from xtb_broker.utils import get_today_ini, get_today_end
from xtb_broker.xtb import Xtb


class Arbitrage(Frozen):
    def __init__(self,
                 xtb: Xtb,
                 name: str,
                 active: bool,
                 tp: float,
                 sl: float,
                 positions: List[Dict]):
        self.__name: str = name
        self.__active: bool = active
        self.__tp: float = tp
        self.__sl: float = sl
        self.__positions: List[Position] = self.__init_positions(xtb=xtb, setup_positions=positions)
        self.__shift: Optional[Shift] = self.__extract_shift()
        self.__status: str = ""
        self.__profit: float = 0
        self.__action: str = ""

    def __init_positions(self, xtb: Xtb, setup_positions: List[Dict]) -> List[Position]:
        positions = []
        for setup_position in setup_positions:
            if "trade" not in setup_position or not setup_position['trade']:
                setup_position['trade'] = dict(
                    cmd=getattr(Cmd(), setup_position['direction']),
                    volume=setup_position['volume'],
                    symbol=setup_position['symbol'],
                    comment=self.__name)
            positions.append(Position(xtb=xtb, **setup_position))
        return positions

    @property
    def name(self) -> str:
        return self.__name

    @property
    def active(self) -> bool:
        return self.__active

    @property
    def active_on_off(self) -> str:
        return "On" if self.__active else "Off"

    @property
    def tp(self) -> float:
        return self.__tp

    @property
    def sl(self) -> float:
        return self.__sl

    @property
    def shift(self) -> Shift:
        return self.__shift

    @property
    def positions(self) -> List[Position]:
        return self.__positions

    @property
    def action(self) -> str:
        return self.__action

    @action.setter
    def action(self, value: str) -> None:
        self.__action = value

    @property
    def status(self) -> str:
        return self.__status

    @status.setter
    def status(self, value: str) -> None:
        self.__status = value

    @property
    def symbols(self) -> List[Symbol]:
        return list(map(lambda position: position.symbol, self.positions))

    @property
    def orders(self) -> List[int]:
        return list(map(lambda position: position.order, self.positions))

    @property
    def trades(self) -> List[Trade]:
        return list(map(lambda position: position.trade, self.positions))

    @property
    def is_working(self):
        return all(map(lambda p: p.is_working, self.positions))

    @property
    def is_completely_accepted(self):
        accepted = list(map(lambda p: p.trade.accepted, self.positions))
        return sum(accepted) == len(accepted)

    @property
    def is_partially_failed(self):
        accepted = list(map(lambda p: p.trade.accepted, self.positions))
        return 0 < sum(accepted) < len(accepted)

    @property
    def is_completely_failed(self):
        accepted = list(map(lambda p: p.trade.accepted, self.positions))
        return 0 == sum(accepted)

    def __extract_shift(self) -> Optional[Shift]:
        if self.is_working:
            shift = Shift(from_ts=get_today_ini(), to_ts=get_today_end())
            for position in self.__positions:
                shift.from_ts = max([shift.from_ts, position.shift.from_ts])
                shift.to_ts = min([shift.to_ts, position.shift.to_ts])
            return shift

    def __get_trades(self, xtb: Xtb):
        trades = xtb.get_client().get_trade_records(orders=self.orders)
        for position in self.positions:
            trade = list(filter(lambda t: t['order'] == position.order, trades))
            if trade:
                position.trade = Trade(trade_raw=list(trade)[0])
            else:
                logger.debug(f"No trade found for order '{position.order} - {position.symbol.symbol}'")
                raise TradeNotFoundException(f"No trade found for order {position.order}")

    def is_operable(self, closing_threshold: int):
        return (self.shift.to_ts - datetime.now()).total_seconds() > closing_threshold

    @property
    def profit(self) -> float:
        return round(sum(map(lambda p: p.profit, self.positions)), 2)

    @retry(retry_on_exception=retry_if_arbitrage_partially_closed_exception,
           stop_max_attempt_number=5, wait_exponential_multiplier=1000, wait_exponential_max=10000)
    def transact(self, xtb: Xtb, _type: int) -> None:
        keep_opening = True
        for idx, position in enumerate(self.positions):
            if keep_opening:
                if not position.trade.accepted:
                    position.trade.transact(xtb=xtb, _type=_type)
                    self.positions[idx].action = position.trade.transaction.status
                if not position.trade.accepted:
                    keep_opening = False

        if self.is_completely_accepted:
            logger.debug("Arbitrage process was SUCCESSFUL and is COMPLETELY CLOSED")
            self.action = "closed"
        if self.is_partially_failed:
            logger.critical("Arbitrage process has FAILED and is PARTIALLY CLOSED")
            self.action = "PARTIALLY closed"
            raise ArbitragePartiallyClosedException("Arbitrage process has FAILED and is PARCIALLY CLOSED")
        if self.is_completely_failed:
            self.action = "COMPLETELY open"
            logger.warning("Arbitrage process has FAILED and is COMPLETELY OPEN")



from datetime import datetime

from xtb_broker.component.frozen import Frozen


class Shift(Frozen):
    def __init__(self, from_ts: datetime, to_ts: datetime):
        self.__from_ts: datetime = from_ts
        self.__to_ts: datetime = to_ts

    @property
    def from_ts(self) -> datetime:
        return self.__from_ts

    @from_ts.setter
    def from_ts(self, value: datetime) -> None:
        self.__from_ts = value

    @property
    def to_ts(self) -> datetime:
        return self.__to_ts

    @to_ts.setter
    def to_ts(self, value: datetime) -> None:
        self.__to_ts = value



from typing import Dict, Optional
from datetime import datetime

from xtb_broker.component.frozen import Frozen
from xtb_broker.component.constants import WEEKDAY
from xtb_broker.models.timetable import Timetable
from xtb_broker.models.shift import Shift

from xtb_broker.xtb import Xtb


class Tick(Frozen):
    def __init__(self, xtb: Xtb, symbol: str, timestamp: int   = None):
        tick_dict: Dict = xtb.get_client().get_symbol(symbol=symbol)
        self.__symbol: str = tick_dict['symbol']
        self.__ask: float = tick_dict['ask']
        self.__bid: float = tick_dict['bid']
        self.__high: float = tick_dict['high']
        self.__low: float = tick_dict['low']
        self.__timestamp: int = timestamp or tick_dict['timestamp']
        self.__spread: float = tick_dict['spreadRaw']

    @property
    def symbol(self) -> str:
        return self.__symbol

    @property
    def ask(self) -> float:
        return self.__ask

    @property
    def bid(self) -> float:
        return self.__bid

    @property
    def high(self) -> float:
        return self.__high

    @property
    def low(self) -> float:
        return self.__low

    @property
    def timestamp(self) -> int:
        return self.__timestamp

    @property
    def spread(self) -> float:
        return self.__spread



from typing import Dict, Optional
from retrying import retry

from xtb_broker.component.frozen import Frozen
from xtb_broker.component.exception import OrderPendingException, \
    retry_if_order_not_sent_exception, retry_if_order_not_accepted_exception
from xtb_broker.component.constants import Type, CMD, Cmd, STATUS, Status
from xtb_broker.component.logger import logger
from xtb_broker.xtb import Xtb
from xtb_broker.models.transaction import Transaction


class Trade(Frozen):
    def __init__(self, trade_raw: Dict):
        trade_raw['order'] = trade_raw.get('order', 0)
        trade_raw['offset'] = trade_raw.get('offset', 0)
        trade_raw['sl'] = trade_raw.get('sl', 0)
        trade_raw['tp'] = trade_raw.get('tp', 0)
        trade_raw['profit'] = trade_raw.get('profit', 0)
        trade_raw['expiration'] = trade_raw.get('expiration', 0)
        trade_raw['storage'] = trade_raw.get('storage', 0)
        self.__order: Optional[int] = trade_raw['order']
        self.__cmd: int = trade_raw['cmd']
        self.__comment: str = trade_raw['comment']
        self.__offset: Optional[float] = trade_raw['offset']
        self.__sl: Optional[float] = trade_raw['sl']
        self.__tp: Optional[float] = trade_raw['tp']
        self.__symbol: str = trade_raw['symbol']
        self.__volume: float = trade_raw['volume']
        self.__profit: Optional[float] = trade_raw['profit']
        self.__storage: Optional[float] = trade_raw['storage']
        self.__raw: Dict = trade_raw
        self.__transaction: Optional[Transaction] = None

        self.__digits: int = trade_raw['digits']
        self.__commission: float = trade_raw['commission']
        self.__close_price: float = trade_raw['close_price']
        self.__open_price: float = trade_raw['open_price']
        self.__open_time: int = trade_raw['open_time']
        self.__close_time: int = trade_raw['close_time']
        self.__timestamp: int = trade_raw['timestamp']
        self.__closed: bool = trade_raw['closed'] == 'true'

    @property
    def order(self) -> int:
        return self.__order

    @order.setter
    def order(self, value: int) -> None:
        self.__order = value

    @property
    def cmd(self) -> int:
        return self.__cmd

    @property
    def comment(self) -> str:
        return self.__comment

    @property
    def offset(self) -> float:
        return self.__offset

    @property
    def sl(self) -> float:
        return self.__sl

    @property
    def tp(self) -> float:
        return self.__tp

    @property
    def symbol(self) -> str:
        return self.__symbol

    @property
    def volume(self) -> float:
        return self.__volume

    @property
    def raw(self) -> Dict:
        return self.__raw

    @property
    def profit(self) -> float:
        return self.__profit

    @property
    def storage(self) -> float:
        return self.__storage

    @property
    def transaction(self) -> Transaction:
        return self.__transaction

    @property
    def accepted(self) -> bool:
        try:
            return self.transaction.accepted
        except:
            return False

    @property
    def digits(self) -> int:
        return self.__digits

    @property
    def commission(self) -> float:
        return self.__commission

    @property
    def close_price(self) -> float:
        return self.__close_price

    @property
    def open_price(self) -> float:
        return self.__open_price

    @property
    def open_time(self) -> int:
        return self.__open_time

    @property
    def close_time(self) -> int:
        return self.__close_time

    @property
    def timestamp(self) -> int:
        return self.__timestamp

    @property
    def closed(self) -> bool:
        return self.__closed

    def transact(self, xtb: Xtb, _type: int) -> None:
        symbol = xtb.get_client().get_symbol(symbol=self.__symbol)
        if _type == Type.OPEN:
            if self.cmd == Cmd.BUY:
                self.raw['open_price'] = symbol['ask']
            elif self.cmd == Cmd.SELL:
                self.raw['open_price'] = symbol['bid']
        self.send_transaction(xtb=xtb, _type=_type)
        self.check_order_sent(xtb=xtb)

    @retry(retry_on_exception=retry_if_order_not_sent_exception,
           stop_max_attempt_number=5, wait_exponential_multiplier=1000, wait_exponential_max=10000)
    def send_transaction(self, xtb: Xtb, _type: int) -> None:
        # TODO: check lotMin del symbol antes de mandar transaction y devolver una exception
        self.transaction = Transaction(order=xtb.get_client().trade_transaction(trade=self.raw, _type=_type))

    @retry(retry_on_exception=retry_if_order_not_accepted_exception,
           stop_max_attempt_number=5, wait_exponential_multiplier=1000, wait_exponential_max=10000)
    def check_order_sent(self, xtb: Xtb):
        try:
            transaction = xtb.get_client().trade_transaction_status(transacting_order=self.transaction.order)
            self.transaction.status = STATUS[transaction['requestStatus']]
            self.transaction.message = transaction['message']
            self.transaction.custom_comment = transaction['customComment']
            self.transaction.ask = transaction['ask']
            self.transaction.bid = transaction['bid']
            if self.transaction.status == Status.ACCEPTED:
                logger.debug(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' is CLOSED. "
                             f"Transaction order is {self.transaction.order}")
            elif self.transaction.status == Status.PENDING:
                logger.warning(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' is PENDING. "
                               f"Transaction order is {self.transaction}")
                raise OrderPendingException(f"Trade order '{self.order} is PENDING. "
                                            f"Closing order is {self.transaction}")
            elif self.transaction.status == Status.ERROR:
                logger.error(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' has FAILED. "
                             f"Transaction: {self.transaction}")
            elif self.transaction.status == Status.REJECTED:
                logger.error(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' has been REJECTED. "
                             f"Transaction order is {self.transaction}")
        except OrderPendingException:
            self.transaction.status = Status.PENDING
        except Exception as e:
            logger.error(e)



from typing import Dict
from dataclasses import dataclass

# Wrapper name and version
WRAPPER_NAME: str = 'python'
WRAPPER_VERSION: str = '2.5.0'

# XAPI connection
XAPI_ADDRESS: str = 'xapi.xtb.com'
XAPI_PORT_DICT: Dict = dict(DEMO=5124, REAL=5112)

# API inter-command timeout (in ms)
API_MIN_SECONDS_BETWEEN_REQUESTS: float = 0.5

# max connection tries
API_MAX_CONN_TRIES: int = 3
API_READ_TIMEOUT: int = 1120

STATUS: Dict = {
    0: "ERROR",
    1: "PENDING",
    3: "ACCEPTED",
    4: "REJECTED"}


@dataclass(frozen=True)
class Status(object):
    ERROR: str = 'ERROR'
    PENDING: str = 'PENDING'
    ACCEPTED: str = 'ACCEPTED'
    REJECTED: str = 'REJECTED'


WEEKDAY: Dict = {
    1: "monday",
    2: "tuesday",
    3: "wednesday",
    4: "thursday",
    5: "friday",
    6: "saturday",
    7: "sunday"}


@dataclass(frozen=True)
class Weekday(object):
    monday: int = 'monday'
    tuesday: int = 'tuesday'
    wednesday: int = 'wednesday'
    thursday: int = 'thursday'
    friday: int = 'friday'
    saturday: int = 'saturday'
    sunday: int = 'sunday'


@dataclass(frozen=True)
class TransactionSide(object):
    BUY: int = 0
    SELL: int = 1
    BUY_LIMIT: int = 2
    SELL_LIMIT: int = 3
    BUY_STOP: int = 4
    SELL_STOP: int = 5


@dataclass(frozen=True)
class TransactionType(object):
    ORDER_OPEN: int = 0
    ORDER_CLOSE: int = 2
    ORDER_MODIFY: int = 3
    ORDER_DELETE: int = 4


@dataclass(frozen=True)
class Type:
    OPEN: int = 0
    PENDING: int = 1
    CLOSE: int = 2
    MODIFY: int = 3
    DELETE: int = 4


@dataclass(frozen=True)
class LogLevel:
    DEBUG: str = "DEBUG"
    INFO: str = "INFO"
    WARNING: str = "WARNING"
    ERROR: str = "ERROR"
    CRITICAL: str = "CRITICAL"

LOG_LEVEL: str = LogLevel.INFO


@dataclass(frozen=True)
class Cmd:
    BUY: int = 0
    SELL: int = 1


CMD: Dict = {0: "BUY", 1: "SELL"}


@dataclass(frozen=True)
class Period:
    M1: int = 1
    M5: int = 5
    M15: int = 15
    M30: int = 30
    H1: int = 60
    H4: int = 240
    D1: int = 1440
    W1: int = 10080
    MN1: int = 43200



from typing import Optional, Dict
from retrying import retry
from datetime import datetime
from xtb_broker.component.constants import WEEKDAY
from xtb_broker.models.timetable import Timetable
from xtb_broker.models.shift import Shift
from xtb_broker.utils import ts_to_unix_ts
import xtb_broker.client
from xtb_broker.client import XtbClient


# Declared at cold-start, but only initialized if/when the function executes (initializing them lazily on demand)
xtb_client: Optional[XtbClient] = None


class Xtb:
    def __init__(self, mode: str):
        self.__mode: str = mode
        self.__xtb_client: Optional[XtbClient] = None

    def __get_client(
        self,
        _client: Optional[client.XtbClient] = None,
        reconnect: bool = False
    ) -> Optional[client.XtbClient]:
        try:
            if not _client or reconnect:
                client_new: XtbClient = get_xtb_client(self.__mode)
                return client_new
        except Exception as e:
            raise Exception(f'__get_client:: {e}')

    def get_client(self, reconnect: bool = False) -> XtbClient:
        global xtb_client
        xtb_client_new: Optional[XtbClient] = self.__get_client(xtb_client, reconnect)
        if xtb_client_new:
            # create xtb_client -> initializing them lazily on demand
            xtb_client = xtb_client_new
        return xtb_client

    def is_now_server_time(self, server_time: datetime = None) -> bool:
        if server_time is None:
            server_time = self.get_client().get_server_time().replace(microsecond=0)
        now = datetime.now().replace(microsecond=0)
        return abs(now.timestamp() - server_time.timestamp()) < 60

    def get_symbol_timetable(self, symbol: str) -> Timetable:
        timetable = Timetable()
        timetable.__frozen__ = False
        for day, shifts in self.get_client().get_trading_hours([symbol])[symbol].items():
            timetable.__setattr__(
                WEEKDAY[day].lower(), [Shift(from_ts=shift['from_ts'], to_ts=shift['to_ts']) for shift in shifts])
        timetable.__frozen__ = True
        return timetable

    def get_tick_price(self, symbol: str, level: int = 0, timestamp: int = ts_to_unix_ts(datetime.now())) -> Dict:
        return self.get_client().get_tick_prices(symbols=[symbol], level=level, timestamp=timestamp)[0]

    def get_trade_record(self, order: int) -> Dict:
        return self.get_client().get_trade_records(orders=[order])[0]


@retry(stop_max_attempt_number=3, wait_random_min=1000, wait_random_max=2000)
def get_xtb_client(mode: str) -> XtbClient:
    return XtbClient(mode)



from setuptools import setup, find_packages
import io
import os
import re

VERSION = "0.0.15"
DESCRIPTION = "XTB broker models, methods and tools"
LONG_DESCRIPTION = DESCRIPTION

setup(
    name="xtb_broker",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="Arrubo",
    url="https://pypi.org/project/xtb-broker/",
    packages=find_packages()
)
